import json
from enum import Enum
from pprint import pprint

import requests
from base64 import b64decode, b64encode
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5 as Cipher_PKCS1_v1_5


class APIRequest:

    def __init__(self, context=None):
        self.context = context

    def execute(self):
        if self.context is not None:
            self.create_default_headers()
            # pprint(self.context)
            try:
                return {
                    APIMethodType.GET: self.__get,
                    APIMethodType.POST: self.__post,
                    APIMethodType.PUT: self.__put
                }.get(self.context.method_type, self.__unknown)()
            except requests.exceptions.ConnectionError as ce:
                print(ce)
                return None
        else:
            raise TypeError('Context cannot be None.')

    def create_bearer_token(self):
        key_der = b64decode(self.context.public_key)
        key_pub = RSA.importKey(key_der)
        cipher = Cipher_PKCS1_v1_5.new(key_pub)
        cipher_text = cipher.encrypt(self.context.api_key.encode('ascii'))
        encrypted_msg = b64encode(cipher_text)

        return encrypted_msg

    def create_default_headers(self):
        self.context.add_header('Authorization', 'Bearer {}'.format(
            self.create_bearer_token().decode('utf-8')))
        self.context.add_header('Content-Type', 'application/json')
        self.context.add_header('Host', self.context.address)

    def __get(self):
        r = requests.get(self.context.get_url(
        ), params=self.context.get_parameters(), headers=self.context.get_headers())
        print(r)
        return APIResponse(r.status_code, json.loads(r.headers.__str__().replace("'", '"')), json.loads(r.text))

    def __post(self):
        r = requests.post(self.context.get_url(
        ), headers=self.context.get_headers(), json=self.context.get_parameters())
        print(r)
        return APIResponse(r.status_code, json.loads(r.headers.__str__().replace("'", '"')), json.loads(r.text))

    def __put(self):
        print('PUT')
        r = requests.put(self.context.get_url(
        ), headers=self.context.get_headers(), json=self.context.get_parameters())
        print('PUT', r)
        return APIResponse(r.status_code, json.loads(r.headers.__str__().replace("'", '"')), json.loads(r.text))

    def __unknown(self):
        raise Exception('Unknown Method')


class APIResponse(dict):

    def __init__(self, status_code, headers, body):
        super(APIResponse, self).__init__()
        self['status_code']: str = status_code
        self['headers']: dict = headers
        self['body']: dict = body

    @property
    def status_code(self) -> int:
        return self['status_code']

    @status_code.setter
    def status_code(self, status_code: int):
        if type(status_code) is not int:
            raise TypeError('status_code must be a int')
        else:
            self['status_code'] = status_code

    @property
    def headers(self) -> dict:
        return self['headers']

    @headers.setter
    def headers(self, headers: dict):
        if type(headers) is not dict:
            raise TypeError('headers must be a dict')
        else:
            self['headers'] = headers

    @property
    def body(self) -> dict:
        return self['body']

    @body.setter
    def body(self, body: dict):
        if type(body) is not dict:
            raise TypeError('body must be a dict')
        else:
            self['body'] = body


class APIMethodType(Enum):
    GET: int = 0
    POST: int = 1
    PUT: int = 3
    DELETE: int = 4


class APIContext(dict):

    def __init__(self, api_key='', public_key='', ssl=False, method_type=APIMethodType.GET, address='', port=80, path='', headers={}, parameters={}):
        super(APIContext, self).__init__()

        self['api_key']: str = api_key
        self['public_key']: str = public_key
        self['ssl']: bool = ssl
        self['method_type']: Enum = method_type
        self['address']: str = address
        self['port']: int = port
        self['path']: str = path
        self['headers']: dict = headers
        self['parameters']: dict = parameters

    def get_url(self):
        if self.ssl is True:
            return 'https://{}:{}{}'.format(self.address, self.port, self.path)
        else:
            return 'http://{}:{}{}'.format(self.address, self.port, self.path)

    def add_header(self, header, value):
        self['headers'].update({header: value})

    def get_headers(self):
        return self['headers']

    def add_parameter(self, key, value):
        self['parameters'].update({key: value})

    def get_parameters(self):
        return self['parameters']

    @property
    def api_key(self) -> str:
        return self['api_key']

    @api_key.setter
    def api_key(self, api_key: str):
        if type(api_key) is not str:
            raise TypeError('api_key must be a str')
        else:
            self['api_key'] = api_key

    @property
    def public_key(self) -> str:
        return self['public_key']

    @public_key.setter
    def public_key(self, public_key: str):
        if type(public_key) is not str:
            raise TypeError('public_key must be a str')
        else:
            self['public_key'] = public_key

    @property
    def ssl(self) -> bool:
        return self['ssl']

    @ssl.setter
    def ssl(self, ssl: bool):
        if type(ssl) is not bool:
            raise TypeError('ssl must be a bool')
        else:
            self['ssl'] = ssl

    @property
    def method_type(self) -> APIMethodType:
        return self['method_type']

    @method_type.setter
    def method_type(self, method_type: APIMethodType):
        if type(method_type) is not APIMethodType:
            raise TypeError('method_type must be a APIMethodType')
        else:
            self['method_type'] = method_type

    @property
    def address(self) -> str:
        return self['address']

    @address.setter
    def address(self, address: str):
        if type(address) is not str:
            raise TypeError('address must be a str')
        else:
            self['address'] = address

    @property
    def port(self) -> int:
        return self['port']

    @port.setter
    def port(self, port: int):
        if type(port) is not int:
            raise TypeError('port must be a int')
        else:
            self['port'] = port

    @property
    def path(self) -> str:
        return self['path']

    @path.setter
    def path(self, path: str):
        if type(path) is not str:
            raise TypeError('path must be a str')
        else:
            self['path'] = path


def execute_mpesa_c2b(config, payload):
    api_context = APIContext()
    api_context.api_key = config['api_key']
    api_context.public_key = config['public_key']
    api_context.ssl = True
    api_context.method_type = APIMethodType.POST
    api_context.address = 'api.sandbox.vm.co.mz'
    api_context.port = 18352
    api_context.path = '/ipg/v1x/c2bPayment/singleStage/'

    api_context.add_header('Origin', '*')

    api_context.add_parameter(
        'input_TransactionReference', payload['reference'])
    api_context.add_parameter('input_CustomerMSISDN', payload['msisdn'])
    api_context.add_parameter('input_Amount', payload['amount'])
    api_context.add_parameter(
        'input_ThirdPartyReference', payload['third_party_reference'])
    api_context.add_parameter(
        'input_ServiceProviderCode', payload['service_provider_code'])

    api_request = APIRequest(api_context)
    result = api_request.execute()

    return result.body


def execute_mpesa_reversal(config, payload):
    api_context = APIContext()
    api_context.api_key = config['api_key']
    api_context.public_key = config['public_key']
    api_context.ssl = True
    api_context.method_type = APIMethodType.PUT
    api_context.address = 'api.sandbox.vm.co.mz'
    api_context.port = 18354
    api_context.path = '/ipg/v1x/reversal/'

    api_context.add_header('Origin', '*')

    api_context.add_parameter('input_TransactionID', payload['transaction'])
    api_context.add_parameter('input_SecurityCredential', 'Mpesa2019')
    api_context.add_parameter('input_InitiatorIdentifier', 'MPesa2018')
    api_context.add_parameter(
        'input_ThirdPartyReference', payload['third_party_reference'])
    api_context.add_parameter(
        'input_ServiceProviderCode', payload['service_provider_code'])
    api_context.add_parameter('input_ReversalAmount', payload['amount'])

    api_request = APIRequest(api_context)
    result = api_request.execute()

    return result.body



from django.utils.translation import gettext_lazy
from . import __version__

try:
    from pretix.base.plugins import PluginConfig
except ImportError:
    raise RuntimeError("Please use pretix 2.7 or above to run this plugin!")


class PluginApp(PluginConfig):
    default = True
    name = "pretix_mpesamz"
    verbose_name = "MPesaMZ"

    class PretixPluginMeta:
        name = gettext_lazy("MPesaMZ")
        author = "Ivan Ruby"
        description = gettext_lazy("A plugin to allow payments through the M-Pesa mobile payment platform")
        visible = True
        version = __version__
        category = "PAYMENT"
        compatibility = "pretix>=2.7.0"

    def ready(self):
        from . import signals  # NOQA





import json
from pprint import pprint
from collections import OrderedDict

from i18nfield.fields import I18nFormField, I18nTextarea
from i18nfield.strings import LazyI18nString

from pretix.base.models import OrderPayment, OrderRefund
from pretix.base.payment import BasePaymentProvider, PaymentException
from pretix.base.templatetags.rich_text import rich_text
from django.utils.translation import ugettext_lazy as _
from django.template.loader import get_template
from django import forms

from random import randint
from pretix_mpesamz.mpesa_api import execute_mpesa_c2b
from pretix_mpesamz.mpesa_api import execute_mpesa_reversal


class MpesaPayment(BasePaymentProvider):
    identifier = 'mpesa'

    ########################################################
    #                   General Settings
    ########################################################
    @property
    def verbose_name(self):
        return str(self.settings.get('method_name', as_type=LazyI18nString) or _('M-Pesa'))

    @property
    def confirm_button_name(self):
        return _('Pagar via M-Pesa')

    @property
    def priority(self):
        return 100

    ########################################################
    #                Control Panel Settings
    ########################################################
    @property
    def information_text(self):
        return rich_text(self.settings.get('information_text', as_type=LazyI18nString))

    @property
    def payment_pending_text(self):
        return rich_text(self.settings.get('payment_pending_text', as_type=LazyI18nString))

    @property
    def payment_completed_text(self):
        return rich_text(self.settings.get('payment_completed_text', as_type=LazyI18nString))

    @property
    def settings_form_fields(self):
        service_provider_code_field = forms.CharField(
            label=_('Código de Provedor de Serviço M-Pesa'),
            help_text=_('Identificador da entidade fornecido pelo M-Pesa')
        )

        api_key_field = forms.CharField(
            label=_('API Key'),
            help_text=_('API Key fornecido pelo M-Pesa')
        )

        public_key_field = forms.CharField(
            label=_('Public Key'),
            help_text=_('Public Key fornecido pelo M-Pesa')
        )

        info_field = I18nFormField(
            label=_('Payment information text'),
            help_text=_('Shown to the user when selecting a payment method.'),
            widget=I18nTextarea,
        )
        pending_field = I18nFormField(
            label=_('Payment pending text'),
            help_text=_(
                'Shown to the user when viewing a pending payment order.'),
            widget=I18nTextarea,
        )
        completed_field = I18nFormField(
            label=_('Payment completed text'),
            help_text=_(
                'Shown to the user when viewing an order with completed payment.'),
            widget=I18nTextarea,
        )

        settingsList = [
            ('service_provider_code', service_provider_code_field),
            ('api_key', api_key_field),
            ('public_key', public_key_field),
            ('information_text', info_field),
            ('payment_pending_text', pending_field),
            ('payment_completed_text', completed_field)
        ]

        return OrderedDict(list(super().settings_form_fields.items()) + settingsList)

    ########################################################
    #               Checkout process settings
    ########################################################
    def payment_is_valid_session(self, request):
        return all([
            request.session.get('payment_%s_msisdn' %
                                self.identifier, '') != ''
        ])

    def order_change_allowed(self, order):
        return True

    ########################################################
    #                   General Settings
    ########################################################

    @property
    def payment_form_fields(self):
        # Validate: ensure only valid numbers can be entered
        msisdn_field = ('msisdn',
                        forms.IntegerField(
                            label='Número (+258):',
                            required=True,
                            min_value=840000000,
                            max_value=859999999,
                            error_messages={
                                'invalid': 'Por favor, introduza um número 84 ou 85 válido'}
                        ))
        return OrderedDict([
            msisdn_field,
        ])

    def payment_form_render(self, request):

        form = self.payment_form(request)
        template = get_template('pretix_mpesamz/checkout_payment_form.html')
        ctx = {
            'request': request,
            'form': form,
            'information_text': self.information_text,
        }
        return template.render(ctx)

    def checkout_confirm_render(self, request):
        template = get_template('pretix_mpesamz/order.html')
        ctx = {
            'information_text': self.information_text,
            'msisdn': request.session.get('payment_%s_msisdn' % self.identifier)
        }
        return template.render(ctx)

    def execute_payment(self, request, payment):
        msisdn = request.session.get('payment_%s_msisdn' % self.identifier, '')

        try:
            config = {
                'api_key': self.settings.get('api_key'),
                'public_key': self.settings.get('public_key')
            }

            payment_data = {
                'msisdn': "258" + str(msisdn),
                'reference': 'TKT' + str(randint(0, 10000)),
                'third_party_reference': payment.order.code,
                'amount': str(int(float(payment.amount))),
                'service_provider_code': self.settings.get('service_provider_code')
            }

            result = execute_mpesa_c2b(config, payment_data)
            # pprint(result)

            if result['output_ResponseCode'] == 'INS-0':
                success_payload = json.dumps({
                    'success': True,
                    'code': result['output_ResponseCode'],
                    'msisdn': msisdn,
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })
                # Displayed in control panel but not on API order response
                payment.info = success_payload

                payment.save(update_fields=['info'])
                payment.confirm()
            else:
                error_payload = json.dumps({
                    'success': False,
                    'code': result['output_ResponseCode'],
                    'msisdn': msisdn,
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })

                # Displayed in control panel but not on API order response
                payment.info = error_payload

                payment.save(update_fields=['info'])
                payment.fail()

            return
        except Exception as e:
            print(repr(e))

    def payment_prepare(self, request, payment):
        msisdn = request.session.get('payment_%s_msisdn' % self.identifier, '')

        try:
            config = {
                'api_key': self.settings.get('api_key'),
                'public_key': self.settings.get('public_key')
            }

            payment_data = {
                'msisdn': "258" + str(msisdn),
                'reference': 'TKT' + str(randint(0, 10000)),
                'third_party_reference': payment.order.code + str(randint(1, 10)),
                'amount': str(int(float(payment.amount))),
                'service_provider_code': self.settings.get('service_provider_code')
            }

            result = execute_mpesa_c2b(config, payment_data)
            # pprint(result)

            if result['output_ResponseCode'] == 'INS-0':
                success_payload = json.dumps({
                    'success': True,
                    'code': result['output_ResponseCode'],
                    'msisdn': msisdn,
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })
                # Displayed in control panel but not on API order response
                payment.info = success_payload

                payment.save(update_fields=['info'])
                payment.confirm()
            else:
                error_payload = json.dumps({
                    'success': False,
                    'code': result['output_ResponseCode'],
                    'msisdn': msisdn,
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })

                # Displayed in control panel but not on API order response
                payment.info = error_payload

                payment.save(update_fields=['info'])
                payment.fail()

            return
        except Exception as e:
            print(repr(e))

    def payment_pending_render(self, request, payment) -> str:
        template = get_template('pretix_mpesamz/payment_pending.html')
        if payment.info is not None and 'conversation' in payment.info:
            payment_info = json.loads(payment.info)
        else:
            return _("No payment information available.")

        if payment_info['code'] == "INS-6":
            reason = "PIN incorreto"
        elif payment_info['code'] == "INS-9":
            reason = "demora na comunicação entre a aplicação e o M-Pesa"
        elif payment_info['code'] == "INS-10":
            reason = "transação duplicada"
        else:
            reason = payment_info['description']

        ctx = {
            'reason': reason,
            'msisdn': request.session.get('payment_%s_msisdn' % self.identifier, '')
        }
        return template.render(ctx)

    def order_completed_render(self, request, order) -> str:
        template = get_template('pretix_mpesamz/order.html')
        if order.payment_info:
            payment_info = json.loads(order.payment_info)
        else:
            return _("No payment information available.")
        ctx = {
            'information_text': self.payment_completed_text,
            'msisdn': self.msisdn
        }
        return template.render(ctx)

    ########################################################
    #                 Refund payment for Order
    ########################################################
    def payment_refund_supported(self, payment: OrderPayment):
        return True

    def payment_partial_refund_supported(self, payment: OrderPayment):
        return False

    def execute_refund(self, refund: OrderRefund):
        try:
            # pprint(vars(refund.payment))
            payment_info = json.loads(refund.payment.info)
            config = {
                'api_key': self.settings.get('api_key'),
                'public_key': self.settings.get('public_key')
            }

            payment_data = {
                'transaction': payment_info['transaction'],
                'third_party_reference': refund.order.code,
                'amount': str(refund.amount),
                'service_provider_code': self.settings.get('service_provider_code')
            }

            result = execute_mpesa_reversal(config, payment_data)

            pprint(result)

            if result['output_ResponseCode'] == 'INS-0':
                success_payload = json.dumps({
                    'success': True,
                    'code': result['output_ResponseCode'],
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })
                # Displayed in control panel but not on API order response
                refund.payment.info = success_payload
                refund.info = success_payload
                refund.done()
            else:
                error_payload = json.dumps({
                    'success': False,
                    'code': result['output_ResponseCode'],
                    'conversation': result['output_ConversationID'],
                    'description': result['output_ResponseDesc'],
                    'transaction': result['output_TransactionID']
                })
                refund.payment.info = error_payload
                refund.info = error_payload

                raise PaymentException(error_payload['description'])

        except Exception as e:
            print(repr(e))
            raise PaymentException('M-Pesa refund could not be completed')

    def refund_control_render(self, request, refund) -> str:
        template = get_template('pretix_mpesamz/refund_control.html')

        if refund.info is not None and 'transaction' in refund.info:
            refund_info = json.loads(refund.info)
        else:
            return _("Devolução via M-Pesa sem sucesso.")

        ctx = {
            'description': refund_info['description'],
            'code': refund_info['code'],
            'transaction': refund_info['transaction']
        }
        return template.render(ctx)

    ########################################################
    #                   Called to display Order
    #                   info in Control Panel
    ########################################################

    def payment_control_render(self, request, payment) -> str:
        template = get_template('pretix_mpesamz/control.html')

        if payment.info is not None and 'conversation' in payment.info:
            payment_info = json.loads(payment.info)
        else:
            return _("Pagamento via M-Pesa sem sucesso.")

        ctx = {
            'conversationID': payment_info['conversation'],
            'msisdn': payment_info['msisdn'],
            'description': payment_info['description'],
            'code': payment_info['code'],
            'transaction': payment_info['transaction']
        }
        return template.render(ctx)



from django.dispatch import receiver

from pretix.base.signals import register_payment_providers
from pretix.presale.signals import order_meta_from_request

@receiver(register_payment_providers, dispatch_uid="payment_mpesa.provider")
def register_payment_provider(sender, **kwargs):
    from .payment import MpesaPayment
    return MpesaPayment


import os
from distutils.command.build import build

from django.core import management
from setuptools import find_packages, setup

from pretix_mpesamz import __version__


try:
    with open(
        os.path.join(os.path.dirname(__file__), "README.rst"), encoding="utf-8"
    ) as f:
        long_description = f.read()
except Exception:
    long_description = ""


class CustomBuild(build):
    def run(self):
        management.call_command("compilemessages", verbosity=1)
        build.run(self)


cmdclass = {"build": CustomBuild}


setup(
    name="pretix-mpesa-mz",
    version=__version__,
    description="A plugin to allow payments through the M-Pesa mobile payment platform",
    long_description=long_description,
    url="https://github.com/ivanruby/pretix-mpesa-mz",
    author="Ivan Ruby",
    author_email="ivanrubyds@gmail.com",
    license="Apache",
    install_requires=[],
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    cmdclass=cmdclass,
    entry_points="""
[pretix.plugin]
pretix_mpesamz=pretix_mpesamz:PretixPluginMeta
""",
)
